from enum import StrEnum
from logging import getLogger
from typing import Generic, TypeVar

from pydantic import BaseModel, Field
from unique_toolkit.tools.config import get_configuration_dict

logger = getLogger(__name__)


class ContentProcessingStartegy(StrEnum):
    SUMMARIZE = "summarize"
    TRUNCATE = "truncate"


T = TypeVar("T", bound=ContentProcessingStartegy)


class ContentProcessingStrategyConfig(BaseModel, Generic[T]):
    model_config = get_configuration_dict()

    strategy: T = Field(
        description="The strategy to use for content processing",
    )



from typing import Callable

from pydantic import BaseModel, Field
from unique_toolkit.tools.config import get_configuration_dict

from unique_web_search.services.preprocessing.cleaning.patterns import (
    PATTERNS,
    get_cleaner,
)


class CleaningStrategyConfig(BaseModel):
    model_config = get_configuration_dict()

    markdown_cleaning_timeout: float = Field(
        default=5.0,
        description="Timeout for markdown cleaning.",
    )
    remove_nested_images_and_links: bool = Field(
        default=False,
        description="Whether to clean nested images and links in the content.",
    )
    remove_simple_images_and_links: bool = Field(
        default=False,
        description="Whether to clean simple images and links in the content.",
    )
    remove_multiple_linebreaks: bool = Field(
        default=False,
        description="Whether to clean multiple linebreaks in the content.",
    )
    remove_repeating_patterns: bool = Field(
        default=False,
        description="Whether to clean repeating patterns in the content.",
    )


def _encoding_cleanup(text: str) -> str:
    """Fast encoding cleanup - most expensive operations."""
    text = text.encode("utf-8", "ignore").decode()
    return text


def get_cleaning_strategy(
    config: CleaningStrategyConfig,
) -> Callable[[str], str]:
    timeout = config.markdown_cleaning_timeout
    fun_stack = []
    if config.remove_nested_images_and_links:
        fun_stack.append(get_cleaner(PATTERNS.NESTED_IMAGES_AND_LINKS, timeout))
    if config.remove_simple_images_and_links:
        fun_stack.append(get_cleaner(PATTERNS.IMAGES_AND_LINKS, timeout))
    if config.remove_multiple_linebreaks:
        fun_stack.append(get_cleaner(PATTERNS.MULTIPLE_LINEBREAKS, timeout))
    if config.remove_repeating_patterns:
        fun_stack.append(get_cleaner(PATTERNS.REPEATING_PATTERNS, timeout))

    fun_stack.append(_encoding_cleanup)

    def cleaning_function(text: str) -> str:
        for fn in fun_stack:
            text = fn(text)
        return text

    return cleaning_function



import asyncio
import logging

import pandas as pd
from langchain.text_splitter import TokenTextSplitter
from pydantic import BaseModel, Field
from unidecode import unidecode
from unique_toolkit import LanguageModelService
from unique_toolkit.app.schemas import ChatEvent
from unique_toolkit.content.schemas import ContentChunk
from unique_toolkit.embedding.service import EmbeddingService
from unique_toolkit.language_model.infos import LanguageModelInfo
from unique_toolkit.tools.config import get_configuration_dict
from unique_toolkit.tools.utils.execution.execution import SafeTaskExecutor

from unique_web_search.services.preprocessing.content_processing import (
    ProcessingStrategyType,
    get_strategy,
)
from unique_web_search.services.preprocessing.content_processing.strategies import (
    TruncatePageToMaxTokensConfig,
)
from unique_web_search.services.preprocessing.schema import (
    WebPageChunk,
)
from unique_web_search.services.search_engine.schema import (
    WebSearchResult,
)

logger = logging.getLogger(__name__)


DEFAULT_ENCODER_MODEL = "cl100k_base"


class ContentAdapterConfig(BaseModel):
    model_config = get_configuration_dict()
    chunk_size: int = Field(
        default=1000,
        description="Number of chunks to split the search results",
    )

    chunking_max_workers: int = Field(
        default=10,
        description="Number of workers to embed search results",
    )

    content_processing_strategy_config: ProcessingStrategyType = Field(
        default_factory=TruncatePageToMaxTokensConfig,
        description="The strategy to use for content processing",
        discriminator="strategy",
    )


class ContentAdapter:
    def __init__(
        self,
        event: ChatEvent,
        config: ContentAdapterConfig,
        llm_service: LanguageModelService,
        language_model: LanguageModelInfo,
    ):
        self.config = config
        self.embedding_service = EmbeddingService(event=event)
        self.language_model = language_model
        self.encoder_name = language_model.encoder_name or DEFAULT_ENCODER_MODEL
        self.llm_service = llm_service

    async def run(self, query: str, pages: list[WebSearchResult]) -> pd.DataFrame:
        """
        Preprocess the pages content.

        Args:
            pages: list of pages.

        Returns:
            pd.DataFrame: DataFrame with the pages.
        """

        pages = await self._process_pages(query, pages)

        pages_chunks = self._split_pages_to_chunks(pages)

        logger.info(f"Number of chunks total: {len(pages_chunks)}")

        df_chunks = pd.DataFrame([record.model_dump() for record in pages_chunks])

        ###
        # If embedding reranking flag is false, return the chunks
        ###
        df_chunks = self._create_sources(df_chunks)
        return df_chunks

    # TODO: Find a tracking solution
    # @track(
    #     tags=["content_processing"],
    # )
    async def _process_pages(
        self, query: str, pages: list[WebSearchResult]
    ) -> list[WebSearchResult]:
        logger.info(
            f"Processing pages with strategy: {self.config.content_processing_strategy_config.strategy}"
        )
        content_processing_strategy = get_strategy(
            strategy=self.config.content_processing_strategy_config,
            query=query,
            language_model=self.language_model,
            llm_service=self.llm_service,
        )

        safe_task_executor = SafeTaskExecutor(
            logger=logger,
        )

        results = await asyncio.gather(
            *[
                safe_task_executor.execute_async(
                    content_processing_strategy,
                    page=page,
                )
                for page in pages
            ]
        )

        processed_pages = []
        for result, page in zip(results, pages):
            if result.success:
                page = result.unpack()
            else:
                # Empty content to avoid overfilling the context in case processing strategy fails
                page.content = ""
            processed_pages.append(page)

        return processed_pages

    def _split_pages_to_chunks(
        self, pages: list[WebSearchResult]
    ) -> list[WebPageChunk]:
        """Build a vector store for the pages.

        Args:
            pages: list of pages.

        Returns:
            pd.DataFrame: DataFrame with the pages.
        """
        records = [
            record
            for chunks_records in map(self._create_chunks, pages)
            for record in chunks_records
        ]
        return records

    def _create_sources(self, df_chunks: pd.DataFrame):
        """
        Create source info for referencing in stream.

        Args:
            df_chunks: The DataFrame with the chunks.
        """

        if df_chunks.empty:
            return df_chunks

        # Function to create the source dictionary
        def create_source(row):
            display_link = row["display_link"]

            # Convert to ascii
            title = unidecode(row["title"])

            name = f'{display_link}: "{title}"'

            content_chunk = ContentChunk(
                id=name,
                text=row["content"],
                order=row["order"],
                start_page=None,
                end_page=None,
                key=name,
                chunk_id=row["order"],
                url=row["url"],
                title=name,
            )

            return content_chunk.model_dump()

        df_chunks["source"] = df_chunks.apply(create_source, axis=1)

        return df_chunks

    def _create_chunks(self, page: WebSearchResult) -> list[WebPageChunk]:
        """
        Create chunks from the page content.

        Args:
            page: dict with the page content.

        Returns:
            list[dict]: list of chunks.
        """

        chunk_size = self.config.chunk_size

        splitter = TokenTextSplitter(
            encoding_name=self.encoder_name,
            chunk_size=chunk_size,
            chunk_overlap=0,
        )

        chunks = splitter.split_text(page.content)

        if len(chunks) == 0:
            return [
                WebPageChunk(
                    url=page.url,
                    display_link=page.display_link,
                    title=page.title,
                    snippet=page.snippet,
                    content=self._wrap_with_snipper(page.snippet, page.content),
                    order="0",
                )
            ]

        records = [
            WebPageChunk(
                url=page.url,
                display_link=page.display_link,
                title=page.title,
                snippet=page.snippet,
                content=self._wrap_with_snipper(page.snippet, chunk),
                order=str(order),
            )
            for order, chunk in enumerate(chunks)
        ]
        return records

    @staticmethod
    def _wrap_with_snipper(snippet: str, content: str) -> str:
        if len(content) > 0:
            return f"<SearchEngineSnippet>{snippet}</SearchEngineSnippet>\n\n<FetchContent>{content}</FetchContent>"
        return f"<SearchEngineSnippet>{snippet}</SearchEngineSnippet>"



DEFAULT_TOOL_DESCRIPTION_FOR_SYSTEM_PROMPT = (
    "Use the `WebSearch` tool to access up-to-date information from the web or when responding to the user requires information about their location. Some examples of when to use the `WebSearch` tool include:\n"
    "- Local Information: Use the `WebSearch` tool to respond to questions that require information about the user's location, such as the weather, local businesses, or events.\n"
    "- Freshness: If up-to-date information on a topic could potentially change or enhance the answer, call the `WebSearch` tool any time you would otherwise refuse to answer a question because your knowledge might be out of date.\n"
    "- Niche Information: If the answer would benefit from detailed information not widely known or understood (which might be found on the internet), such as details about a small neighborhood, a less well-known company, or arcane regulations, use web sources directly rather than relying on the distilled knowledge from pretraining.\n"
    "- Accuracy: If the cost of a small mistake or outdated information is high (e.g., using an outdated version of a software library or not knowing the date of the next game for a sports team), then use the `WebSearch` tool.\n\n"
    "**Instruction Query Splitting**\n"
    'You should split the user question into multiple queries when the user\'s question needs to be decomposed / rewritten to find different facts. Perform for each query an individual tool call. Avoid short queries that are extremely broad and will return unrelated results. Strip the search string of any extraneous details, e.g. instructions or unnecessary context. However, you must fill in relevant context from the rest of the conversation to make the question complete. E.g. "What was their age?" => "What was Kevin\'s age?" because the preceding conversation makes it clear that the user is talking about Kevin.\n\n'
    "Here are some examples of how to use the WebSearch tool:\n"
    'User: What was the GDP of France and Italy in the 1970s? => queries: ["france gdp 1970", "italy gdp 1970"] # Splitting of the query into 2 queries and perform 2 tool calls\n'
    'User: What does the report say about the GPT4 performance on MMLU? => queries: ["GPT4 performance on MMLU?"] # Simplify the query'
)
DEFAULT_TOOL_FORMAT_INFORMATION_FOR_SYSTEM_PROMPT = (
    "Whenever you use information retrieved with the InternalSearch, you must adhere to strict reference guidelines. "
    "You must strictly reference each fact used with the `source_number` of the corresponding passage, in the following format: '[source<source_number>]'.\n\n"
    "Example:\n"
    "- The stock price of Apple Inc. is $150 [source0] and the company's revenue increased by 10% [source1].\n"
    "- Moreover, the company's market capitalization is $2 trillion [source2][source3].\n"
    "- Our internal documents tell us to invest[source4] (Internal)\n\n"
    "A fact is preferably referenced by ONLY ONE source, e.g [sourceX], which should be the most relevant source for the fact.\n"
    "Follow these guidelines closely and be sure to use the proper `source_number` when referencing facts.\n"
    "Make sure that your reference follow the format [sourceX] and that the source number is correct.\n"
    "Source is written in singular form and the number is written in digits.\n\n"
    "IT IS VERY IMPORTANT TO FOLLOW THESE GUIDELINES!!\n"
    "NEVER CITE A source_number THAT YOU DON'T SEE IN THE TOOL CALL RESPONSE!!!\n"
    "The source_number in old assistant messages are no longer valid.\n"
    "EXAMPLE: If you see [source34] and [source35] in the assistant message, you can't use [source34] again in the next assistant message, this has to be the number you find in the message with role 'tool'.\n"
    "BE AWARE:All tool calls have been filtered to remove uncited sources. Tool calls return much more data than you see.\n\n"
    "### Web Search Answering Protocol for Employee Questions\n"
    "When using web search to assist employees, follow\n"
    "this structured approach to ensure accurate,\n"
    "well-sourced, and relevant responses:\n\n"
    "#### 1. Evaluate and Prioritize Web Sources\n"
    "Give strong preference to:\n"
    "- **Top-ranked results**, especially the first 3–5\n"
    "- **Reliable sources**, such as:\n"
    "  - Established news organizations\n"
    "  - Academic institutions\n"
    "  - Government websites\n"
    "  - Recognized industry authorities\n"
    "- **Recent content**, particularly when the topic requires up-to-date information\n"
    "- **Primary sources** over summaries or interpretations\n\n"
    "#### 2. Source Reliability Guidelines\n"
    "- Give **highest weight** to:\n"
    "  - Peer-reviewed publications\n"
    "  - Trusted media (e.g., *BBC*, *Reuters*, *The New York Times*)\n"
    "  - Official or governmental resources\n"
    "- Be cautious with:\n"
    "  - Blogs, forums, or unaudited websites\n"
    "- Recognize and prioritize **domain expertise** (e.g., medical content from healthcare institutions)\n"
    "- Always consider the **recency** and **relevance** of the content\n\n"
    "#### 3. Acknowledge Limitations\n"
    "- If the search results do not confidently answer the question, **clearly state** this limitation."
)

DEFAULT_TOOL_DESCRIPTION = (
    "Issues a query to a search engine and displays the results from the web. "
    "If user provide a specific link, you should include it in the query (not only the domain)."
)

REFINE_QUERY_SYSTEM_PROMPT = """
You're task consist of a query for a search engine.

** Refine the query Guidelines **
- The query should be a string that does not exceed 6 key words.
- Never include temporal information in the refined query. There is a separate field for this purpose.
- You may add the additional advanced syntax when relevant to refine the results:
- Use quotes `"..."` for exact words (avoid doing it for phrases as it will dramatically reduce the number of results).
- Use `-word` to exclude terms.
- Use `site:domain.com` to restrict to a site.
- Use `intitle:`, `inurl:` to target title/URL.
- Use `OR` for alternatives, `*` as a wildcard.
- Use `..` for number ranges (e.g., 2010..2020).
- Use `AROUND(N)` to find terms close together.
- Use `define:word` for definitions.
- Combine operators for powerful filtering.

** IMPORTANT **
- You should not use any date restriction in the refined query.
""".strip()



from logging import getLogger

from pydantic import BaseModel, Field, model_validator
from pydantic.json_schema import SkipJsonSchema
from unique_toolkit._common.chunk_relevancy_sorter.config import (
    ChunkRelevancySortConfig,
)
from unique_toolkit._common.default_language_model import DEFAULT_GPT_4o
from unique_toolkit._common.feature_flags.schema import (
    FeatureExtendedSourceSerialization,
)
from unique_toolkit._common.validators import LMI, get_LMI_default_field
from unique_toolkit.evals.schemas import EvaluationMetricName
from unique_toolkit.language_model.infos import ModelCapabilities
from unique_toolkit.tools.config import get_configuration_dict
from unique_toolkit.tools.schemas import BaseToolConfig

from unique_web_search.prompts import (
    DEFAULT_TOOL_DESCRIPTION,
    DEFAULT_TOOL_DESCRIPTION_FOR_SYSTEM_PROMPT,
    DEFAULT_TOOL_FORMAT_INFORMATION_FOR_SYSTEM_PROMPT,
    REFINE_QUERY_SYSTEM_PROMPT,
)
from unique_web_search.services.content_adapter import ContentAdapterConfig
from unique_web_search.services.preprocessing.cleaning.cleaning_strategy import (
    CleaningStrategyConfig,
)
from unique_web_search.services.preprocessing.crawlers.basic import BasicCrawlerConfig
from unique_web_search.services.preprocessing.crawlers.crawl4ai import (
    Crawl4AiCrawlerConfig,
)
from unique_web_search.services.preprocessing.crawlers.firecrawl import (
    FirecrawlCrawlerConfig,
)
from unique_web_search.services.preprocessing.crawlers.jina import JinaCrawlerConfig
from unique_web_search.services.preprocessing.crawlers.tavily import TavilyCrawlerConfig
from unique_web_search.services.search_engine.firecrawl import FireCrawlConfig
from unique_web_search.services.search_engine.google import GoogleConfig
from unique_web_search.services.search_engine.jina import JinaConfig
from unique_web_search.services.search_engine.tavily import TavilyConfig

logger = getLogger(__name__)

DEFAULT_MODEL_NAME = DEFAULT_GPT_4o


class AnswerGenerationConfig(BaseModel):
    model_config = get_configuration_dict()
    limit_token_sources: int = Field(
        default=10000,
        description="Token Source Limit",
    )
    max_chunks_to_consider: int = Field(
        default=20,
        description="Token Source Limit",
    )
    number_history_interactions_included: int = Field(
        default=2,
        description="Number of history interactions included",
    )


class ExperimentalFeatures(FeatureExtendedSourceSerialization): ...


class QueryRefinementConfig(BaseModel):
    model_config = get_configuration_dict()
    enabled: bool = Field(
        default=True,
        description="Whether to enable the refined query",
    )
    system_prompt: str = Field(
        default=REFINE_QUERY_SYSTEM_PROMPT,
        description="The system prompt to refine the query",
    )


class WebSearchToolParametersConfig(BaseModel):
    model_config = get_configuration_dict()
    query_description: str = Field(
        default="The search query to issue to the web.",
        description="The tool parameter query description",
    )
    date_restrict_description: str = Field(
        default="""Restricts results to a recent time window. Format: `[period][number]` — `d`=days, `w`=weeks, `m`=months, `y`=years.  
Examples: `d1` (24h), `w1` (1 week), `m3` (3 months), `y1` (1 year).  
Omit for no date filter. Avoid adding date terms in the main query.""",
        description="The tool parameter date restrict description",
    )


class WebSearchConfig(BaseToolConfig):
    language_model: LMI = get_LMI_default_field(DEFAULT_MODEL_NAME)

    limit_token_sources: SkipJsonSchema[int] = Field(
        default=60_000,  # TODO: Remove SkipJsonSchema once UI (Spaces 2.0) can be configured to not include certain fields
        description="Token Source Limit",
    )
    percentage_of_input_tokens_for_sources: float = Field(
        default=0.4,
        description="The percentage of the maximum input tokens of the language model to use for the tool response.",
        ge=0.0,
        le=1.0,
    )
    language_model_max_input_tokens: SkipJsonSchema[int | None] = Field(
        default=None,
        description="Language model maximum input tokens",
    )

    search_engine_config: (
        GoogleConfig
        | SkipJsonSchema[JinaConfig]
        | SkipJsonSchema[TavilyConfig]
        | SkipJsonSchema[FireCrawlConfig]
    ) = Field(
        default_factory=GoogleConfig,
        description="Search Engine Configuration",
        discriminator="search_engine_name",
        title="Search Engine Configuration",
    )

    crawler_config: (
        Crawl4AiCrawlerConfig
        | BasicCrawlerConfig
        | SkipJsonSchema[FirecrawlCrawlerConfig]
        | SkipJsonSchema[JinaCrawlerConfig]
        | SkipJsonSchema[TavilyCrawlerConfig]
    ) = Field(
        default_factory=BasicCrawlerConfig,
        description="The crawler configuration.",
        discriminator="crawler_type",
    )

    cleaning_strategy_config: CleaningStrategyConfig = Field(
        default_factory=CleaningStrategyConfig,
        description="The cleaning strategy configuration",
        title="Cleaning Strategy Configuration",
    )

    content_adapter_config: ContentAdapterConfig = Field(
        default_factory=ContentAdapterConfig,
        description="The content adapter configuration",
        title="Content Adapter Configuration",
    )

    chunk_relevancy_sort_config: ChunkRelevancySortConfig = Field(
        default_factory=ChunkRelevancySortConfig,
        description="Chunk Relevancy Sort Configuration",
        title="Chunk Relevancy Sort Configuration",
    )

    evaluation_check_list: list[EvaluationMetricName] = Field(
        default=[
            EvaluationMetricName.HALLUCINATION,
        ],
        description="Check list of evaluations executed conditionally after the answer is generated",
        title="Evaluation Check List",
    )

    query_refinement_config: QueryRefinementConfig = Field(
        default_factory=QueryRefinementConfig,
        description="The query refinement configuration",
        title="Query Refinement Configuration",
    )

    tool_parameters_config: WebSearchToolParametersConfig = Field(
        default_factory=WebSearchToolParametersConfig,
        description="The tool parameters configuration",
        title="Tool Parameters Configuration",
    )

    tool_description: str = Field(
        default=DEFAULT_TOOL_DESCRIPTION,
        description="The tool description",
    )

    tool_description_for_system_prompt: str = Field(
        default=DEFAULT_TOOL_DESCRIPTION_FOR_SYSTEM_PROMPT,
    )

    tool_format_information_for_system_prompt: str = Field(
        default=DEFAULT_TOOL_FORMAT_INFORMATION_FOR_SYSTEM_PROMPT,
    )

    experimental_features: ExperimentalFeatures = ExperimentalFeatures()

    @model_validator(mode="after")
    def disable_query_refinement_if_no_structured_output(self):
        if ModelCapabilities.STRUCTURED_OUTPUT not in self.language_model.capabilities:
            self.query_refinement_config.enabled = False
            logger.warning(
                "The language model does not support structured output. Query refinement is disabled."
            )
        return self



from logging import getLogger
from typing import Literal

import tiktoken
from pydantic import Field
from unique_toolkit import LanguageModelService
from unique_toolkit.language_model.builder import MessagesBuilder
from unique_toolkit.language_model.infos import (
    LanguageModelInfo,
)

from unique_web_search.services.preprocessing.content_processing.base import (
    ContentProcessingStartegy,
    ContentProcessingStrategyConfig,
)
from unique_web_search.services.search_engine.schema import (
    WebSearchResult,
)

logger = getLogger(__name__)


class SummarizeWebpageConfig(ContentProcessingStrategyConfig):
    strategy: Literal[ContentProcessingStartegy.SUMMARIZE] = (
        ContentProcessingStartegy.SUMMARIZE
    )

    pre_truncate_to_max_tokens: int = Field(
        default=30000,
        description="Max number of tokens to truncate the page to keep before summarization",
    )

    min_tokens_trigger_summarization: int = Field(
        default=5000,
        description="Min number of tokens to trigger summarization",
    )
    summarization_system_prompt: str = Field(
        default="""You are a helping assistant that generates query focused summarization of a webpage content. The summary should convey any information that is relevant to the query.""",
        description="The system prompt to use for summarization",
    )


async def summarize_webpage(
    llm_service: LanguageModelService,
    query: str,
    page: WebSearchResult,
    model_info: LanguageModelInfo,
    pre_truncate_to_max_tokens: int,
    min_tokens_trigger_summarization: int,
    summarization_system_prompt: str,
) -> WebSearchResult:
    page_content, token_count = await truncate_page_to_max_tokens(
        page,
        model_info.encoder_name,
        pre_truncate_to_max_tokens,
    )

    if token_count < min_tokens_trigger_summarization:
        return page

    logger.info(f"Summarizing webpage ({page.url}) with {token_count} tokens")

    messages = (
        MessagesBuilder()
        .system_message_append(summarization_system_prompt)
        .user_message_append(f"Query: {query}\nWebpage: {page_content}".strip())
        .build()
    )

    response = await llm_service.complete_async(
        messages=messages,
        model_name=model_info.name,
    )

    page.content = response.choices[0].message.content  # type: ignore

    return page


class TruncatePageToMaxTokensConfig(ContentProcessingStrategyConfig):
    strategy: Literal[ContentProcessingStartegy.TRUNCATE] = (
        ContentProcessingStartegy.TRUNCATE
    )

    truncate_to_max_tokens: int = Field(
        default=5000,
        description="Max number of tokens to truncate the page to",
    )


async def truncate_page_to_max_tokens(
    page: WebSearchResult,
    encoder_name: str,
    max_tokens: int,
) -> tuple[WebSearchResult, int]:
    encoder = tiktoken.get_encoding(encoder_name)

    tokens = encoder.encode(page.content)

    token_count = len(tokens)

    if token_count > max_tokens:
        page.content = encoder.decode(tokens[:max_tokens])

    return page, token_count


def get_strategy(
    strategy: ContentProcessingStrategyConfig,
    query: str,
    language_model: LanguageModelInfo,
    llm_service: LanguageModelService,
):
    if isinstance(strategy, SummarizeWebpageConfig):

        async def summarize_webpage_wrapper(
            page: WebSearchResult,
        ) -> WebSearchResult:
            return await summarize_webpage(
                page=page,
                model_info=language_model,
                llm_service=llm_service,
                query=query,
                pre_truncate_to_max_tokens=strategy.pre_truncate_to_max_tokens,
                min_tokens_trigger_summarization=strategy.min_tokens_trigger_summarization,
                summarization_system_prompt=strategy.summarization_system_prompt,
            )

        return summarize_webpage_wrapper

    elif isinstance(strategy, TruncatePageToMaxTokensConfig):

        async def truncate_page_to_max_tokens_wrapper(
            page: WebSearchResult,
        ) -> WebSearchResult:
            page, _ = await truncate_page_to_max_tokens(
                page=page,
                encoder_name=language_model.encoder_name,
                max_tokens=strategy.truncate_to_max_tokens,
            )
            return page

        return truncate_page_to_max_tokens_wrapper
    else:
        raise ValueError(f"Invalid strategy: {strategy}")



from unique_web_search.services.preprocessing.crawlers.base import (
    CrawlerType,
)
from unique_web_search.services.preprocessing.crawlers.basic import (
    BasicCrawler,
    BasicCrawlerConfig,
)
from unique_web_search.services.preprocessing.crawlers.crawl4ai import (
    Crawl4AiCrawler,
    Crawl4AiCrawlerConfig,
)
from unique_web_search.services.preprocessing.crawlers.firecrawl import (
    FirecrawlCrawler,
    FirecrawlCrawlerConfig,
)
from unique_web_search.services.preprocessing.crawlers.jina import (
    JinaCrawler,
    JinaCrawlerConfig,
)
from unique_web_search.services.preprocessing.crawlers.tavily import (
    TavilyCrawler,
    TavilyCrawlerConfig,
)

__all__ = [
    "BasicCrawler",
    "BasicCrawlerConfig",
    "Crawl4AiCrawler",
    "Crawl4AiCrawlerConfig",
    "FirecrawlCrawler",
    "FirecrawlCrawlerConfig",
    "JinaCrawler",
    "JinaCrawlerConfig",
    "TavilyCrawler",
    "TavilyCrawlerConfig",
    "CrawlerType",
]



from typing import Callable

import regex
import timeout_decorator


# Pre-compiled regex patterns for better performance
class OptimizedPatterns:
    """Pre-compiled regex patterns for efficient text cleaning."""

    # Handles nested patterns
    NESTED_IMAGES_AND_LINKS = regex.compile(
        r"!\[[^\]]*\]\(.*\)|\[[^\]]*\]\(.*\)", regex.DOTALL
    )

    # Handles non-nested patterns
    IMAGES_AND_LINKS = regex.compile(
        r"!\[[^\]]*\]\([^)]*\)|\[[^\]]*\]\([^)]*\)", regex.DOTALL
    )

    # Multiple linebreaks
    MULTIPLE_LINEBREAKS = regex.compile(r"\n{2,}")

    # Repeating patterns
    REPEATING_PATTERNS = regex.compile(r"(.{5,}?)(?:\s*\1){2,}", regex.MULTILINE)


PATTERNS = OptimizedPatterns()


def get_cleaner(pattern: regex.Pattern, timeout: float = 0.5) -> Callable[[str], str]:
    """Get a function that removes a specific pattern from text."""

    @timeout_decorator.timeout(timeout)
    def cleaner(text: str) -> str:
        return pattern.sub("", text)

    return cleaner



from unique_web_search.services.preprocessing.content_processing.base import (
    ContentProcessingStartegy,
    ContentProcessingStrategyConfig,
)
from unique_web_search.services.preprocessing.content_processing.strategies import (
    SummarizeWebpageConfig,
    TruncatePageToMaxTokensConfig,
    get_strategy,
)

ProcessingStrategyType = SummarizeWebpageConfig | TruncatePageToMaxTokensConfig

__all__ = [
    "ContentProcessingStartegy",
    "ContentProcessingStrategyConfig",
    "get_strategy",
]



def _query_params_to_human_string(query: str, date_restrict: str | None) -> str:
    """
    Converts the WebSearchToolParameters and optional date_restrict to a human-understandable string.
    Maps date_restrict codes to human-readable periods.
    """
    query_str = f"{query}"
    if date_restrict:
        # Map date_restrict codes to human-readable strings
        mapping = {
            "d": "day",
            "w": "week",
            "m": "month",
            "y": "year",
        }
        import re

        match = re.fullmatch(r"([dwmy])(\d+)", date_restrict)
        if match:
            period, number = match.groups()
            period_str = mapping.get(period, period)
            # Pluralize if number > 1
            if number == "1":
                date_str = f"1 {period_str}"
            else:
                date_str = f"{number} {period_str}s"
        else:
            date_str = date_restrict
        query_str += f" (For the last {date_str})"
    return query_str



from datetime import datetime
from time import time

import pandas as pd
from pydantic import BaseModel, ConfigDict, Field, create_model
from typing_extensions import override
from unique_toolkit._common.chunk_relevancy_sorter.service import ChunkRelevancySorter
from unique_toolkit._common.utils.structured_output.schema import StructuredOutputModel
from unique_toolkit.chat.schemas import ChatMessage
from unique_toolkit.chat.service import LanguageModelToolDescription
from unique_toolkit.content.schemas import ContentChunk
from unique_toolkit.content.utils import (
    count_tokens,
)
from unique_toolkit.evals.schemas import EvaluationMetricName
from unique_toolkit.history_manager.utils import transform_chunks_to_string
from unique_toolkit.language_model.builder import MessagesBuilder
from unique_toolkit.language_model.schemas import (
    LanguageModelFunction,
    LanguageModelMessage,
    LanguageModelToolMessage,
)
from unique_toolkit.tools.agent_chunks_hanlder import AgentChunksHandler
from unique_toolkit.tools.factory import ToolFactory
from unique_toolkit.tools.schemas import ToolCallResponse
from unique_toolkit.tools.tool import Tool
from unique_toolkit.tools.tool_progress_reporter import ProgressState

from unique_web_search.config import WebSearchConfig
from unique_web_search.services.content_adapter import ContentAdapter
from unique_web_search.services.search_and_crawl import SearchAndCrawlService
from unique_web_search.services.search_engine.base import SearchEngineType
from unique_web_search.services.search_engine.firecrawl import FireCrawlSearch
from unique_web_search.services.search_engine.google import GoogleSearch
from unique_web_search.services.search_engine.jina import JinaSearch
from unique_web_search.services.search_engine.tavily import TavilySearch
from unique_web_search.utils import _query_params_to_human_string


class RefinedQuery(StructuredOutputModel):
    """A refined query."""

    optimized_query: str = Field(description="The refined query.")


class WebSearchToolParameters(BaseModel):
    """Parameters for the Websearch tool."""

    model_config = ConfigDict(extra="forbid")
    query: str
    date_restrict: str | None

    @classmethod
    def from_tool_parameter_query_description(
        cls, query_description: str, date_restrict_description: str
    ) -> type[BaseModel]:
        """Create a new model with the query field."""
        return create_model(
            cls.__name__,
            query=(str, Field(description=query_description)),
            date_restrict=(
                str | None,
                Field(description=date_restrict_description),
            ),
            __base__=cls,
        )


class WebSearchTool(Tool[WebSearchConfig]):
    name = "WebSearch"

    def __init__(self, configuration: WebSearchConfig, *args, **kwargs):
        super().__init__(configuration, *args, **kwargs)

        self.chunk_relevancy_sorter = ChunkRelevancySorter(self.event)

        self.language_model = self.config.language_model

        self.chat_history_chat_messages: list[ChatMessage] = []
        self.chat_history_token_length = 0
        self.chat_history_language_model_messages: list[LanguageModelMessage] = []
        self.chat_history_list_string: list[str] = []
        self.chat_history_string: str = ""

        self.index = 1000

        self.search_and_crawl_service = SearchAndCrawlService(
            company_id=self.event.company_id,
            search_engine_config=self.config.search_engine_config,
            crawler_config=self.config.crawler_config,
            cleaning_strategy_config=self.config.cleaning_strategy_config,
        )

        self.content_adapter = ContentAdapter(
            event=self.event,
            config=self.config.content_adapter_config,
            llm_service=self._language_model_service,
            language_model=self.config.language_model,
        )

    def _get_search_engine_service(self):
        """Get the search engine service based on the search engine type."""
        search_engine_name = self.config.search_engine_config.search_engine_name

        match search_engine_name:
            case SearchEngineType.JINA:
                return JinaSearch(self.event, self.config.search_engine_config)
            case SearchEngineType.GOOGLE:
                return GoogleSearch(self.event, self.config.search_engine_config)
            case SearchEngineType.FIRECRAWL:
                return FireCrawlSearch(self.event, self.config.search_engine_config)
            case SearchEngineType.TAVILY:
                return TavilySearch(self.event, self.config.search_engine_config)

    @override
    def tool_description(self) -> LanguageModelToolDescription:
        return LanguageModelToolDescription(
            name=self.name,
            description=self.config.tool_description,
            parameters=WebSearchToolParameters.from_tool_parameter_query_description(
                self.config.tool_parameters_config.query_description,
                self.config.tool_parameters_config.date_restrict_description,
            ),
        )

    def tool_description_for_system_prompt(self) -> str:
        return self.config.tool_description_for_system_prompt

    def tool_format_information_for_system_prompt(self) -> str:
        return self.config.tool_format_information_for_system_prompt

    def evaluation_check_list(self) -> list[EvaluationMetricName]:
        return self.config.evaluation_check_list

    def get_evaluation_checks_based_on_tool_response(
        self,
        tool_response: ToolCallResponse,
    ) -> list[EvaluationMetricName]:
        evaluation_check_list = self.evaluation_check_list()

        # Check if the tool response is empty
        if not tool_response.content_chunks:
            return []
        return evaluation_check_list

    async def run(self, tool_call: LanguageModelFunction) -> ToolCallResponse:
        """Run search and generate answer to the user query"""
        self.logger.info("Running the WebSearch tool")

        debug_info = {}

        self.event.payload.user_message.text = (
            self.event.payload.user_message.text or ""
        )

        ###
        # 1. Translate the user query to the search engine query
        ###
        start_time_step_1 = time()
        parameters = WebSearchToolParameters.model_validate(
            tool_call.arguments,
        )

        search_query = parameters.query
        date_restrict = parameters.date_restrict
        debug_info["search_query"] = search_query
        debug_info["date_restrict"] = date_restrict

        # Refine the query to be more specific and relevant to the user's question.
        try:
            if self.config.query_refinement_config.enabled:
                refined_query = await self._refine_query(search_query)
                debug_info["refined_query"] = refined_query.optimized_query
                search_query = refined_query.optimized_query
        except Exception:
            self.logger.exception(
                "An expected error occurred while refining the query. Query refinement is skipped."
            )

        query_human_string = _query_params_to_human_string(search_query, date_restrict)

        if self.tool_progress_reporter:
            await self.tool_progress_reporter.notify_from_tool_call(
                tool_call=tool_call,
                name="**Web search**",
                message=f"{query_human_string}",
                state=ProgressState.RUNNING,
            )

        self.logger.debug(f"Debug info: {debug_info}")

        # debug_info["question_language"] = language
        end_time_step_1 = time()
        self.logger.info(
            f"Step 1: Translate the user query to the search engine query took {end_time_step_1 - start_time_step_1} seconds",
        )

        ###
        # 2. Search the web for relevant information
        ###
        start_time_step_2 = time()
        try:
            (
                search_results,
                time_info,
            ) = await self.search_and_crawl_service.search_and_crawl(
                query=search_query,
                date_restrict=date_restrict,
            )
            debug_info["search_results"] = [
                elem.model_dump() for elem in search_results
            ]
            debug_info["time_info"] = time_info
        except Exception as e:
            self.logger.error(f"An error occurred while searching the web: {e}")
            return ToolCallResponse(
                id=tool_call.id,  # type: ignore
                name=self.name,
                content_chunks=[],
                error_message=str(e),
                debug_info=debug_info,
            )

        end_time_step_2 = time()
        self.logger.info(
            f"Step 2: Search the web for relevant information took {end_time_step_2 - start_time_step_2} seconds",
        )

        try:
            df_relevant_chunks = await self.content_adapter.run(
                query=search_query,
                pages=search_results,
            )
        except Exception as e:
            self.logger.error(
                f"An error occurred while analyzing relevant web pages: {e}"
            )
            return ToolCallResponse(
                id=tool_call.id,  # type: ignore
                name=self.name,
                content_chunks=[],
                error_message=str(e),
                debug_info=debug_info,
            )

        ###
        # 3. Select relevant sources
        ###
        start_time_step_3 = time()
        if self.tool_progress_reporter:
            await self.tool_progress_reporter.notify_from_tool_call(
                tool_call=tool_call,
                name="**Web search**",
                message=f"{query_human_string} (_Postprocessing search results_)",
                state=ProgressState.RUNNING,
            )
        if df_relevant_chunks.empty:
            self.logger.info("No relevant sources found")
            return ToolCallResponse(
                id=tool_call.id,  # type: ignore
                name=self.name,
                content_chunks=[],
                debug_info=debug_info,
            )
        relevant_sources = await self._select_relevant_sources(
            search_query,
            df_relevant_chunks,
            tool_call,
        )
        debug_info["num chunks in final prompts"] = len(relevant_sources)

        end_time_step_3 = time()
        self.logger.info(
            f"Step 3: Select relevant sources took {end_time_step_3 - start_time_step_3} seconds",
        )

        if self.tool_progress_reporter:
            await self.tool_progress_reporter.notify_from_tool_call(
                tool_call=tool_call,
                name="**Web search**",
                message=f"{query_human_string}",
                state=ProgressState.FINISHED,
            )

        return ToolCallResponse(
            id=tool_call.id,  # type: ignore
            name=self.name,
            content_chunks=relevant_sources,
            debug_info=debug_info,
        )

    async def _select_relevant_sources(
        self,
        search_query: str,
        df_chunks: pd.DataFrame,
        tool_call: LanguageModelFunction,
    ) -> list[ContentChunk]:
        # Reduce the sources to the token limit defined in the config
        top_results = self._reduce_sources_to_token_limit(df_chunks)

        # Bring sources into ContentChunk format
        content = [
            ContentChunk.model_validate(source) for source in top_results["source"]
        ]

        # Apply chunk relevancy sorting
        if self.config.chunk_relevancy_sort_config.enabled:
            if self.tool_progress_reporter:
                await self.tool_progress_reporter.notify_from_tool_call(
                    tool_call=tool_call,
                    name="**Web search**",
                    message=f"{search_query} (_Resorting {len(content)} search results_ 🔄)",
                    state=ProgressState.RUNNING,
                )
            sorted_chunks = await self.chunk_relevancy_sorter.run(
                input_text=search_query,
                chunks=content,
                config=self.config.chunk_relevancy_sort_config,
            )
            self.logger.info(f"Sorting chunks message: {sorted_chunks.user_message}")
            return sorted_chunks.content_chunks
        else:
            return content

    def _get_max_tokens(self) -> int:
        if self.config.language_model_max_input_tokens is not None:
            max_tokens = int(
                self.config.language_model_max_input_tokens
                * self.config.percentage_of_input_tokens_for_sources
            )
            self.logger.debug(
                "Using %s of max tokens %s as token limit: %s",
                self.config.percentage_of_input_tokens_for_sources,
                self.config.language_model_max_input_tokens,
                max_tokens,
            )
            return max_tokens
        else:
            self.logger.debug(
                "language model input context size is not set, using default max tokens"
            )
            return (
                min(
                    self.config.limit_token_sources,
                    self.language_model.token_limits.token_limit_input - 1000,
                )
                if self.language_model.token_limits
                and self.language_model.token_limits.token_limit_input
                else self.config.limit_token_sources
            )

    def _reduce_sources_to_token_limit(self, df: pd.DataFrame) -> pd.DataFrame:
        total_tokens = 0
        row_count = 0

        max_token_sources = self._get_max_tokens()

        for i, source in df.iterrows():
            if total_tokens < max_token_sources - self.chat_history_token_length:
                total_tokens += count_tokens(text=source.content)
                row_count += 1
            else:
                break

        return df.head(row_count)

    def get_tool_call_result_for_loop_history(
        self,
        tool_response: ToolCallResponse,
        agent_chunks_handler: AgentChunksHandler,
    ) -> LanguageModelMessage:
        """Process the results of the tool.

        Args:
        ----
            tool_response: The tool response.
            loop_history: The loop history.

        Returns:
        -------
            The tool result to append to the loop history.

        """
        self.logger.debug(
            f"Appending tool call result to history: {tool_response.name}"
        )
        # Initialize content_chunks if None
        content_chunks = tool_response.content_chunks or []

        # Get the maximum source number in the loop history
        max_source_number = len(agent_chunks_handler.chunks)

        # Transform content chunks into sources to be appended to tool result
        sources = transform_chunks_to_string(
            content_chunks,
            max_source_number,
            None,  # Use None for SourceFormatConfig
            self.config.experimental_features.full_sources_serialize_dump,
        )

        # Append the result to the history
        return LanguageModelToolMessage(
            content=sources,
            tool_call_id=tool_response.id,  # type: ignore
            name=tool_response.name,
        )

    # TODO: How to do tracking nicely
    # @track(
    #     tags=["web_search_tool", "refine_query"],
    # )
    async def _refine_query(self, query: str) -> RefinedQuery:
        """Refine the query to be more specific and relevant to the user's question."""
        messages = (
            MessagesBuilder()
            .system_message_append(self.config.query_refinement_config.system_prompt)
            .user_message_append(
                f"Current date: {datetime.now().strftime('%Y-%m-%d')}"
                f"Refine the following query: {query}"
            )
            .build()
        )
        response = await self.language_model_service.complete_async(
            messages,
            model_name=self.language_model.name,
            structured_output_model=RefinedQuery,
            structured_output_enforce_schema=True,
        )

        parsed_response = response.choices[0].message.parsed
        if parsed_response is None:
            raise ValueError("Failed to parse insights from LLM response")

        return RefinedQuery.model_validate(parsed_response)


ToolFactory.register_tool(WebSearchTool, WebSearchConfig)
